import inspect

from opentelemetry.sdk.resources import Resource
from opentelemetry.semconv.resource import ResourceAttributes
from pydantic import create_model
from pydantic_settings import BaseSettings

ResourceAttributesModel = create_model(
    "ResourceAttributesModel",
    **{
        name: (str | None, None)
        for name, value in inspect.getmembers(
            ResourceAttributes, predicate=lambda obj: isinstance(obj, str)
        )
        if name not in ["SCHEMA_URL", "__module__"]
    },
)


class EZResource(ResourceAttributesModel, BaseSettings):
    """
    Construct Resource using `env`

        for example `SERVICE_NAME(service.name)`
        EZ_RESOURCE_SERVICE_NAME=ez-otlp
    """

    schema_url: str | None = None

    class Config:
        env_prefix = "EZ_RESOURCE_"
        case_sensitive = False

    def get_resource(self) -> Resource:
        return Resource.create(
            {
                getattr(ResourceAttributes, k): v
                for k, v in self.model_dump(exclude_none=True).items()
            },
            self.schema_url,
        )


ez_resource = EZResource()
default_resource: Resource = ez_resource.get_resource()

if __name__ == "__main__":
    print(ez_resource.model_dump(exclude_none=True))
    print(default_resource.to_json())



"""https://www.structlog.org/"""

from dataclasses import dataclass, field
from typing import Callable

import logging
from datetime import datetime, timezone

import orjson
from opentelemetry._logs import std_to_otel
from opentelemetry.sdk._logs._internal import LoggerProvider, LogRecord
from opentelemetry.trace import get_current_span

import structlog
from structlog._log_levels import NAME_TO_LEVEL
from structlog.types import BindableLogger, Processor, WrappedLogger


@dataclass
class StructLoggingConfig:
    processors: list[Processor] | None = field(default=None)
    wrapper_class: type[BindableLogger] | None = field(default=None)
    logger_factory: Callable[..., WrappedLogger] | None = field(default=None)
    cache_logger_on_first_use: bool = field(default=True)

    def configure(self):
        structlog.configure(
            processors=self.processors,
            wrapper_class=self.wrapper_class,
            logger_factory=self.logger_factory,
            cache_logger_on_first_use=self.cache_logger_on_first_use,
        )
        return structlog.get_logger


_EXCLUDE_ATTRS = {"exception", "timestamp"}

_dumps = orjson.dumps


class StructlogHandler:
    def __init__(self, logger_provider: LoggerProvider) -> None:
        self._logger = logger_provider.get_logger(__name__)

    def __call__(
        self,
        logger: structlog.typing.WrappedLogger,
        name: str,
        event_dict: structlog.typing.EventDict,
    ):
        span_context = get_current_span().get_span_context()
        severity_number = std_to_otel(NAME_TO_LEVEL[event_dict["level"]])
        attributes = {
            k: v if isinstance(v, (bool, str, bytes, int, float)) else _dumps(v)
            for k, v in event_dict.items()
            if k not in _EXCLUDE_ATTRS
        }

        self._logger.emit(
            LogRecord(
                timestamp=int(datetime.now(timezone.utc).timestamp() * 1e9),
                trace_id=span_context.trace_id,
                span_id=span_context.span_id,
                trace_flags=span_context.trace_flags,
                severity_text=event_dict["level"],
                severity_number=severity_number,
                body=event_dict["event"],
                resource=self._logger.resource,
                attributes=attributes,
            )
        )
        return event_dict


def get_struct_logging_config(
    logger_provider: LoggerProvider,
    log_lever: int = logging.DEBUG,
) -> StructLoggingConfig:
    return StructLoggingConfig(
        processors=[
            structlog.contextvars.merge_contextvars,
            structlog.processors.add_log_level,
            structlog.processors.format_exc_info,
            structlog.processors.TimeStamper(fmt="iso", utc=True),
            StructlogHandler(logger_provider),
            structlog.processors.JSONRenderer(serializer=orjson.dumps),
        ],
        logger_factory=structlog.BytesLoggerFactory(),
        wrapper_class=structlog.make_filtering_bound_logger(log_lever),
        cache_logger_on_first_use=True,
    )



from typing import Optional, Sequence, Union, Tuple, Dict, Literal

from grpc import ChannelCredentials, Compression
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import (
    OTLPLogExporter as gRPCLogExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
    OTLPMetricExporter as gRPCMetricExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as gRPCTraceExporter,
)
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as HttpLogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as HttpMetricExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HttpTraceExporter,
)
from opentelemetry.sdk._logs.export import LogExporter
from opentelemetry.sdk.metrics._internal.aggregation import Aggregation
from opentelemetry.sdk.metrics.export import AggregationTemporality
from opentelemetry.sdk.metrics.export import MetricExporter
from opentelemetry.sdk.trace.export import SpanExporter
from pydantic_settings import BaseSettings


class Exporter(BaseSettings):
    endpoint: Optional[str] = None
    gRPC: bool = True
    # TODO<Martin>: credentials compatible env settings
    credentials: Optional[ChannelCredentials] = None
    headers: Optional[Union[Sequence[Tuple[str, str]], Dict[str, str], str]] = None
    timeout: Optional[int] = None
    compression: Literal[0, 1, 2] = 0  # 0 NoCompression 1 Deflate 2 Gzip

    class Config:
        env_prefix = "EZ_"
        case_sensitive = False

    def get_span_exporter(
        self, span_exporter_cls: Union[gRPCTraceExporter | HttpTraceExporter] = None
    ) -> SpanExporter:
        self.compression: Compression
        if span_exporter_cls is None:
            span_exporter_cls = gRPCTraceExporter if self.gRPC else HttpTraceExporter
        return span_exporter_cls(
            endpoint=self.endpoint,
            headers=self.headers,
            timeout=self.timeout,
            compression=self.compression,
        )

    def get_log_export(
        self, log_exporter_cls: gRPCLogExporter | HttpLogExporter = None
    ) -> LogExporter:
        self.compression: Compression
        if log_exporter_cls is None:
            log_exporter_cls = gRPCLogExporter if self.gRPC else HttpLogExporter
        return log_exporter_cls(
            endpoint=self.endpoint,
            headers=self.headers,
            timeout=self.timeout,
            compression=self.compression,
        )

    def get_metric_exporter(
        self,
        metric_exporter_cls: Union[gRPCMetricExporter, HttpMetricExporter] = None,
        preferred_temporality: Dict[type, AggregationTemporality] = None,
        preferred_aggregation: Dict[type, Aggregation] = None,
        max_export_batch_size: Optional[int] = None,
    ) -> MetricExporter:
        self.compression: Compression
        if metric_exporter_cls is None:
            metric_exporter_cls = (
                gRPCMetricExporter if self.gRPC else HttpMetricExporter
            )
        return metric_exporter_cls(
            endpoint=self.endpoint,
            headers=self.headers,
            timeout=self.timeout,
            compression=self.compression,
            preferred_temporality=preferred_temporality,
            preferred_aggregation=preferred_aggregation,
        )


if __name__ == "__main__":
    print(Exporter(credentials=ChannelCredentials("test")).model_dump())



import logging
from typing import Union, Type, Optional, Dict, List

from grpc import ChannelCredentials
from opentelemetry import _logs
from opentelemetry import metrics
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import (
    OTLPLogExporter as gRPCLogExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
    OTLPMetricExporter as gRPCMetricExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as gRPCTraceExporter,
)
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as HttpLogExporter,
)
from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
    OTLPMetricExporter as HttpMetricExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HttpTraceExporter,
)
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk._logs._internal.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics._internal.aggregation import Aggregation
from opentelemetry.sdk.metrics.export import AggregationTemporality
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    SimpleSpanProcessor,
)

from ez_otlp.open_telemetry.env_exporter import Exporter
from ez_otlp.open_telemetry.env_resource import EZResource

SpanProcessor = Union[
    SimpleSpanProcessor,
    BatchSpanProcessor,
]


class EZBase:
    exporter: Exporter = None
    resource: Resource = None

    def __init__(
        self,
        exporter: Exporter = None,
        resource: Resource | EZResource = None,
        credentials: Optional[ChannelCredentials] = None,
    ):
        if not self.exporter:
            if not exporter:
                exporter = Exporter(credentials=credentials)
            self.exporter = exporter

        if not self.resource:
            if not resource:
                resource = EZResource()
            if isinstance(resource, EZResource):
                resource: Resource = resource.get_resource()
            self.resource = resource


class EZTracer(EZBase):
    tracer_provider: TracerProvider

    def __init__(
        self,
        exporter: Exporter = None,
        resource: Resource | EZResource = None,
        credentials: Optional[ChannelCredentials] = None,
        /,
        set_tracer_provider: bool = True,
        span_processor_cls: Type[SpanProcessor] = BatchSpanProcessor,
        span_exporter_cls: Union[gRPCTraceExporter | HttpTraceExporter] = None,
    ):
        EZBase.__init__(self, exporter, resource, credentials)

        self.tracer_provider = TracerProvider(resource=self.resource)
        if set_tracer_provider:
            trace.set_tracer_provider(self.tracer_provider)
        self.tracer_provider.add_span_processor(
            span_processor_cls(
                span_exporter=self.exporter.get_span_exporter(
                    span_exporter_cls=span_exporter_cls
                )
            )
        )


class EZMetric(EZBase):
    meter_provider: MeterProvider

    def __init__(
        self,
        exporter: Exporter = None,
        resource: Resource | EZResource = None,
        credentials: Optional[ChannelCredentials] = None,
        /,
        metric_exporter_cls: Union[gRPCMetricExporter, HttpMetricExporter] = None,
        preferred_temporality: Dict[type, AggregationTemporality] = None,
        preferred_aggregation: Dict[type, Aggregation] = None,
        max_export_batch_size: Optional[int] = None,
        set_meter_provider: bool = True,
    ):
        EZBase.__init__(self, exporter, resource, credentials)
        metric_exporter = self.exporter.get_metric_exporter(
            metric_exporter_cls=metric_exporter_cls,
            preferred_temporality=preferred_temporality,
            preferred_aggregation=preferred_aggregation,
            max_export_batch_size=max_export_batch_size,
        )
        self.meter_provider = MeterProvider(
            resource=self.resource,
            metric_readers=[
                PeriodicExportingMetricReader(
                    exporter=metric_exporter,
                ),
            ],
            views=[],
        )
        if set_meter_provider:
            metrics.set_meter_provider(self.meter_provider)


class EZLog(EZBase):
    logger_provider: LoggerProvider

    def __init__(
        self,
        exporter: Exporter = None,
        resource: Resource | EZResource = None,
        credentials: Optional[ChannelCredentials] = None,
        /,
        log: str | List[str] = "logging",
        log_lever: int = logging.DEBUG,
        log_exporter_cls: gRPCLogExporter | HttpLogExporter = None,
        set_logger_provider: bool = True,
    ):
        EZBase.__init__(self, exporter, resource, credentials)

        # 配置日志跟踪
        self.logger_provider = LoggerProvider(resource=self.resource)
        if set_logger_provider:
            _logs.set_logger_provider(self.logger_provider)
        self.logger_provider.add_log_record_processor(
            BatchLogRecordProcessor(
                exporter=self.exporter.get_log_export(log_exporter_cls=log_exporter_cls)
            )
        )
        if isinstance(log, str):
            log = [log]
        if "logging" in log:
            from opentelemetry.sdk._logs import LoggingHandler

            logging.getLogger().addHandler(
                LoggingHandler(level=log_lever, logger_provider=self.logger_provider)
            )
        if "structlog" in log:
            from ez_otlp.log.structlog_config import get_struct_logging_config

            config = get_struct_logging_config(
                log_lever=log_lever, logger_provider=self.logger_provider
            )
            config.configure()


class EZ_OTLP(EZLog, EZMetric, EZTracer, EZBase):
    def __init__(
        self,
        exporter: Exporter = None,
        resource: Resource | EZResource = None,
        credentials: Optional[ChannelCredentials] = None,
        /,
        set_tracer_provider: bool = True,
        span_processor_cls: Type[SpanProcessor] = BatchSpanProcessor,
        span_exporter_cls: Union[gRPCTraceExporter | HttpTraceExporter] = None,
        metric_exporter_cls: Union[gRPCMetricExporter, HttpMetricExporter] = None,
        preferred_temporality: Dict[type, AggregationTemporality] = None,
        preferred_aggregation: Dict[type, Aggregation] = None,
        max_export_batch_size: Optional[int] = None,
        set_meter_provider: bool = True,
        log: str | List[str] = "logging",
        log_lever: int = logging.DEBUG,
        log_exporter_cls: gRPCLogExporter | HttpLogExporter = None,
        set_logger_provider: bool = True,
    ):
        EZBase.__init__(self, exporter, resource, credentials)
        EZTracer.__init__(
            self,
            set_tracer_provider=set_tracer_provider,
            span_processor_cls=span_processor_cls,
            span_exporter_cls=span_exporter_cls,
        )

        EZMetric.__init__(
            self,
            metric_exporter_cls=metric_exporter_cls,
            preferred_temporality=preferred_temporality,
            preferred_aggregation=preferred_aggregation,
            max_export_batch_size=max_export_batch_size,
            set_meter_provider=set_meter_provider,
        )

        EZLog.__init__(
            self,
            log=log,
            log_lever=log_lever,
            log_exporter_cls=log_exporter_cls,
            set_logger_provider=set_logger_provider,
        )
